-- (c) David Cunningham 2009, Licensed under the MIT license: http://www.opensource.org/licenses/mit-license.php

local d3d9 = get_rendersystem() == "Direct3D9 Rendering Subsystem"
local lang = d3d9 and "hlsl" or "glsl"

local function prog(name, lang, type)
        local p = get_gpuprog(name) or make_gpuprog(name, lang, type)
        if p.language ~= lang then
                p:destroy()
                p = make_gpuprog(name, lang, type)
        end
        return p
end

local function process_tex_name (mname, tname)
        local substring = tname:sub(1,2)
        if substring == "./" then
                return (string.gsub(mname, "[^/]+$", tname:sub(3)))
        end
        if substring == "//" then
                return mname.."."..tname:sub(3)
        end
        return tname
end

local function get_colour (v, r, g, b, a)
        if v==nil then return r,g,b,a end
        if type(v)=="number" then
                r = v % 256
                v = (v - r)/256
                g = v % 256
                v = (v - g)/256
                b = v % 256
                v = (v - b)/256
                a = v % 256
                return r/255, g/255, b/255, a/255
        else
                return v[1] or r, v[2] or g, v[3] or b, v[4] or a
        end
end

local function set_texture_filtering (mat, index)
        if settings.filtering then
                if settings.anisotropy == 0 then
                        mat:setTextureFiltering(0,0,index,"LINEAR","LINEAR","LINEAR")
                else
                        mat:setTextureAnisotropy(0,0,index, settings.anisotropy)
                        mat:setTextureFiltering(0,0,index,"ANISOTROPIC","ANISOTROPIC","LINEAR")
                end
        else
                mat:setTextureFiltering(0,0,index,"POINT","POINT","NONE")
        end
end

local function shader_caster_names (dm)
        if dm then
                return "caster_T_f", "caster_T_v"
        else
                return "caster_t_f", "caster_t_v"
        end
end

local function shader_names (dm, nm, sm, stipple)
        if sm then 
                if dm then
                        if nm then
                                if stipple then
                                        return "TNSS_f", "TNSS_v"
                                else
                                        return "TNSs_f", "TNSs_v"
                                end
                        else
                                if stipple then
                                        return "TnSS_f", "TnSS_v"
                                else
                                        return "TnSs_f", "TnSs_v"
                                end
                        end
                else
                        if nm then
                                if stipple then
                                        return "tNSS_f", "tNSS_v"
                                else
                                        return "tNSs_f", "tNSs_v"
                                end
                        else
                                if stipple then
                                        return "tnSS_f", "tnSS_v"
                                else
                                        return "tnSs_f", "tnSs_v"
                                end
                        end
                end
        else
                if dm then
                        if nm then
                                if stipple then
                                        return "TNsS_f", "TNsS_v"
                                else
                                        return "TNss_f", "TNss_v"
                                end
                        else
                                if stipple then
                                        return "TnsS_f", "TnsS_v"
                                else
                                        return "Tnss_f", "Tnss_v"
                                end
                        end
                else
                        if nm then
                                if stipple then
                                        return "tNsS_f", "tNsS_v"
                                else
                                        return "tNss_f", "tNss_v"
                                end
                        else
                                if stipple then
                                        return "tnsS_f", "tnsS_v"
                                else
                                        return "tnss_f", "tnss_v"
                                end
                        end
                end
        end
end

local function make_program_caster_cg (diffuse_map)
        local fname, vname = shader_caster_names(diffuse_map)
        local defines = "-O3 -DUSE_DIFFUSE_MAP="..(diffuse_map and "1" or "0")
                      .." -DUSE_STIPPLE_TEXTURE=1 -DNUM_SHADOW_MAPS=0"
        local tunit = 0
        if true then
                defines = defines .. " -DTEXUNIT_STIPPLE=s"..tunit
                tunit = tunit+1
        end
        if diffuse_map then
                defines = defines .. " -DTEXUNIT_DIFFUSE=s"..tunit
                tunit = tunit+1
        end

        local vp = prog(vname,"cg","VERTEX")
        vp.profiles = {"vs_3_0", "vp40"}
        vp.sourceFile = "system/uber.cg"
        vp.entryPoint = "caster_vp_main"
        vp.compileArguments = defines
        vp:reload()
        vp:setAutoConstantInt("world","WORLD_MATRIX")
        vp:setAutoConstantInt("view_proj","VIEWPROJ_MATRIX")
        vp:setAutoConstantInt("light_pos_ws","LIGHT_POSITION")

        local fp = prog(fname,"cg","FRAGMENT")
        fp.profiles = {"ps_3_0", "fp40"}
        fp.sourceFile = "system/uber.cg"
        fp.entryPoint = "caster_fp_main"
        fp.compileArguments = defines
        fp:reload()
        fp:setAutoConstantInt("visibility", "CUSTOM", 0)
        fp:setConstantFloat("alpha_rej", 0)
        fp:setAutoConstantInt("viewport_size","VIEWPORT_SIZE")
        return vp, fp
end

local function make_program_caster (diffuse_map)
        local fname, vname = shader_caster_names(diffuse_map)
        local defines = ";USE_DIFFUSE_MAP="..(diffuse_map and "1" or "0")
                      ..";USE_STIPPLE_TEXTURE=1;NUM_SHADOW_MAPS=0"
        if d3d9 then
                -- for some reason defining GRIT_GLSL didn't work -- seemed macro was not visible
                -- maybe something to do with #version
                defines = defines .. ";GRIT_HLSL=1"
        end
        defines = defines..""

        local vp = prog(vname,lang,"VERTEX")
        vp.target = "vs_3_0"
        vp.sourceFile = "system/uber.shader"
        vp.defines = defines..";CASTER_VERTEX_PROGRAM=1"
        vp.entryPoint = "main"
        vp:reload()
        vp:setAutoConstantInt("world","WORLD_MATRIX")
        vp:setAutoConstantInt("view_proj","VIEWPROJ_MATRIX")
        vp:setAutoConstantInt("light_pos_ws","LIGHT_POSITION")

        local fp = prog(fname,lang,"FRAGMENT")
        fp.target = "ps_3_0"
        fp.sourceFile = "system/uber.shader"
        local tunit = 0
        if true then
                defines = defines .. ";SEM_TEX_STIPPLE=SEM_TEX_"..tunit
                tunit = tunit+1
        end
        if diffuse_map then
                defines = defines .. ";SEM_TEX_DIFFUSE=SEM_TEX_"..tunit
                tunit = tunit+1
        end
        fp.defines = defines..";CASTER_FRAGMENT_PROGRAM=1"
        fp.entryPoint = "main"
        fp:reload()
        if not d3d9 then
                tunit = 0
                if true then
                        fp:setConstantInt("stipple_map", tunit);
                        tunit = tunit+1
                end
                if diffuse_map then
                        fp:setConstantInt("diffuse_map", tunit);
                        tunit = tunit+1
                end
        end
        fp:setAutoConstantInt("visibility", "CUSTOM", 0)
        fp:setConstantFloat("alpha_rej", 0)
        return vp, fp
end

local function material_caster (name, original, tab)

        local shf, shv = shader_caster_names(tab.diffuseMap)

        local mat = get_material(name)
        if mat==nil then
                mat = Material(name)
        else
                -- removing the technique here causes a crash because
                -- the best supported technique no-longer exists.
                mat:removeAllPasses(0)
                mat:createPass(0)
        end

        mat:setFragmentProgram(0,0,shf)
        mat:setVertexProgram(0,0,shv)

        mat:setFragmentConstantFloat(0,0,"alpha_rej", tab.alphaReject or 63/255)

        if tab.backfaces then
                mat:setCull(0,0,false)
                mat:setManualCull(0,0,false)
        end     

        local tex_index = 0

        -- stipple texture
        if true then
                mat:createTextureUnitState(0,0)
                mat:setTextureName(0,0,tex_index, "system/stipple.bmp")
                mat:setTextureFiltering(0,0,tex_index,"POINT","POINT","NONE")
                tex_index = tex_index + 1
        end

        -- diffuse map
        if tab.diffuseMap then
                mat:createTextureUnitState(0,0)
                mat:setTextureName(0,0,tex_index, process_tex_name(original,tab.diffuseMap))
                tex_index = tex_index + 1
        end

        return mat

end



local function make_program_cg (diffuse_map, normal_map, spec_map, stipple)
        local fname, vname = shader_names(diffuse_map, normal_map, spec_map, stipple)
        local defines = "-O3 -DUSE_VERTEX_COLOURS="..(settings.vertexColours and "1" or "0")
                      .." -DUSE_DIFFUSE_MAP="..(diffuse_map and "1" or "0")
                      .." -DUSE_NORMAL_MAP="..(normal_map and "1" or "0")
                      .." -DUSE_SPECULAR_MAP="..(spec_map and "1" or "0")
                      .." -DUSE_FOG="..(settings.fog and "1" or "0")
                      .." -DNO_VERTEX_PROCESSING="..(settings.noVertexProcessing and "1" or "0")
                      .." -DNO_FRAGMENT_PROCESSING="..(settings.noFragmentProcessing and "1" or "0")
                      .." -DUSE_STIPPLE_TEXTURE="..(stipple and "1" or "0")

        if settings.shadow and settings.shadowReceive then
                defines = defines .. " -DNUM_SHADOW_MAPS="..settings.shadowPCSSSplits
                                  .. " -DSHADOW_RES="..settings.shadowRes
                                  .. " -DSHADOW_PADDING="..settings.shadowPCSSPadding
                                  .. " -DSHADOW_DIST1="..settings.shadowPCSSEnd0
                                  .. " -DSHADOW_DIST2="..settings.shadowPCSSEnd1
                                  .. " -DSHADOW_DIST3="..settings.shadowPCSSEnd2
                                  .. " -DSHADOW_FADE_START="..settings.shadowFadeStart
                                  .. " -DSHADOW_FADE_END="..settings["shadowPCSSEnd"..(settings.shadowPCSSSplits-1)]
                                  .. " -DUSE_BILINEAR_PCF_TAP="..(settings.shadowPCFBilinear and "1" or "0")
                                  .. " -DPCF_SIDE="..settings.shadowPCFTaps
                                  .. " -DFAKE_PCF_SIDE="..(settings.shadowPCFEdgeDither and "2" or "1")
                                  .. " -DSPREAD1="..(settings.shadowPCSSSpreadFactor0*settings.shadowPCFSpread)
                                  .. " -DSPREAD2="..(settings.shadowPCSSSpreadFactor1*settings.shadowPCFSpread)
                                  .. " -DSPREAD3="..(settings.shadowPCSSSpreadFactor2*settings.shadowPCFSpread)
        else
                defines = defines .. " -DNUM_SHADOW_MAPS=0"
        end
        local tunit = 0
        if settings.shadow and settings.shadowReceive then
                for i=1,settings.shadowPCSSSplits do
                        defines = defines .. " -DTEXUNIT_SHADOW"..i.."=s"..tunit
                        tunit = tunit+1
                end
        end
        if stipple then
                defines = defines .. " -DTEXUNIT_STIPPLE=s"..tunit
                tunit = tunit+1
        end
        if diffuse_map then
                defines = defines .. " -DTEXUNIT_DIFFUSE=s"..tunit
                tunit = tunit+1
        end
        if normal_map then
                defines = defines .. " -DTEXUNIT_NORMAL=s"..tunit
                tunit = tunit+1
        end
        if spec_map then
                defines = defines .. " -DTEXUNIT_SPECULAR=s"..tunit
                tunit = tunit+1
        end


        local vp = prog(vname,"cg","VERTEX")
        vp.profiles = {"vs_3_0", "vp40"}
        vp.sourceFile = "system/uber.cg"
        vp.compileArguments = defines
        vp.entryPoint = "receiver_vp_main"
        vp:reload()
        vp:setAutoConstantInt("world","WORLD_MATRIX")
        vp:setAutoConstantInt("view_proj","VIEWPROJ_MATRIX")
        if settings.shadow and settings.shadowReceive then
                for i=1,settings.shadowPCSSSplits do
                        vp:setAutoConstantInt("shadow_world_view_proj"..i,"TEXTURE_WORLDVIEWPROJ_MATRIX",i-1)
                end
        end
        vp:setAutoConstantInt("camera_pos","CAMERA_POSITION_OBJECT_SPACE")
        vp:setAutoConstantInt("light_pos_os","LIGHT_POSITION_OBJECT_SPACE",0)
        vp:setAutoConstantInt("light_pos_ws","LIGHT_POSITION")
        vp:setAutoConstantInt("surf_ambient","SURFACE_AMBIENT_COLOUR")
        vp:setAutoConstantInt("surf_emissive","SURFACE_EMISSIVE_COLOUR")
        vp:setAutoConstantInt("light_ambient","AMBIENT_LIGHT_COLOUR")

        local fp = prog(fname,"cg","FRAGMENT")
        fp.profiles = {"ps_3_0", "fp40"}
        fp.sourceFile = "system/uber.cg"
        fp.compileArguments = defines
        fp.entryPoint = "receiver_fp_main"
        fp:reload()
        if settings.fog then
        fp:setAutoConstantInt("the_fog_params", "FOG_PARAMS")
        fp:setAutoConstantInt("the_fog_colour", "FOG_COLOUR")
        end
        fp:setAutoConstantInt("visibility", "CUSTOM", 0)
        fp:setConstantFloat("alpha_rej", 0)
        fp:setAutoConstantInt("light_diffuse", "DERIVED_LIGHT_DIFFUSE_COLOUR", 0)
        fp:setAutoConstantInt("light_specular", "DERIVED_LIGHT_SPECULAR_COLOUR", 0)
        fp:setAutoConstantInt("surf_diffuse", "SURFACE_DIFFUSE_COLOUR")
        fp:setAutoConstantInt("surf_specular", "SURFACE_SPECULAR_COLOUR")
        fp:setAutoConstantInt("surf_shininess", "SURFACE_SHININESS")
        fp:setAutoConstantInt("viewport_size","VIEWPORT_SIZE")


        return vp, fp
end

local function make_program (diffuse_map, normal_map, spec_map, stipple)
        local fname, vname = shader_names(diffuse_map, normal_map, spec_map, stipple)
        local defines = "USE_VERTEX_COLOURS="..(settings.vertexColours and "1" or "0")
                      ..";USE_DIFFUSE_MAP="..(diffuse_map and "1" or "0")
                      ..";USE_NORMAL_MAP="..(normal_map and "1" or "0")
                      ..";USE_SPECULAR_MAP="..(spec_map and "1" or "0")
                      ..";USE_FOG="..(settings.fog and "1" or "0")
                      ..";NO_VERTEX_PROCESSING="..(settings.noVertexProcessing and "1" or "0")
                      ..";NO_FRAGMENT_PROCESSING="..(settings.noFragmentProcessing and "1" or "0")
                      ..";USE_STIPPLE_TEXTURE="..(stipple and "1" or "0")

        if settings.shadow and settings.shadowReceive then
                defines = defines .. ";NUM_SHADOW_MAPS="..settings.shadowPCSSSplits
                                  .. ";SHADOW_RES="..settings.shadowRes
                                  .. ";SHADOW_PADDING="..settings.shadowPCSSPadding
                                  .. ";SHADOW_DIST1="..settings.shadowPCSSEnd0
                                  .. ";SHADOW_DIST2="..settings.shadowPCSSEnd1
                                  .. ";SHADOW_DIST3="..settings.shadowPCSSEnd2
                                  .. ";SHADOW_FADE_START="..settings.shadowFadeStart
                                  .. ";SHADOW_FADE_END="..settings["shadowPCSSEnd"..(settings.shadowPCSSSplits-1)]
                                  .. ";USE_BILINEAR_PCF_TAP="..(settings.shadowPCFBilinear and "1" or "0")
                                  .. ";PCF_SIDE="..settings.shadowPCFTaps
                                  .. ";FAKE_PCF_SIDE="..(settings.shadowPCFEdgeDither and "2" or "1")
                                  .. ";SPREAD1="..(settings.shadowPCSSSpreadFactor0*settings.shadowPCFSpread)
                                  .. ";SPREAD2="..(settings.shadowPCSSSpreadFactor1*settings.shadowPCFSpread)
                                  .. ";SPREAD3="..(settings.shadowPCSSSpreadFactor2*settings.shadowPCFSpread)
        else
                defines = defines .. ";NUM_SHADOW_MAPS=0"
        end
        if d3d9 then
                -- for some reason defining GRIT_GLSL didn't work -- seemed macro was not visible
                -- maybe something to do with #version
                defines = defines .. ";GRIT_HLSL=1"
        end

        local vp = prog(vname,lang,"VERTEX")
        vp.target = "vs_3_0"
        vp.sourceFile = "system/uber.shader"
        vp.defines = defines..";VERTEX_PROGRAM=1"
        vp.entryPoint = "main"
        vp:reload()
        vp:setAutoConstantInt("world","WORLD_MATRIX")
        vp:setAutoConstantInt("view_proj","VIEWPROJ_MATRIX")
        if settings.shadow and settings.shadowReceive then
                for i=1,settings.shadowPCSSSplits do
                        vp:setAutoConstantInt("shadow_world_view_proj"..i,"TEXTURE_WORLDVIEWPROJ_MATRIX",i-1)
                end
        end
        vp:setAutoConstantInt("camera_pos","CAMERA_POSITION_OBJECT_SPACE")
        vp:setAutoConstantInt("light_pos_os","LIGHT_POSITION_OBJECT_SPACE",0)
        vp:setAutoConstantInt("light_pos_ws","LIGHT_POSITION")
        vp:setAutoConstantInt("surf_ambient","SURFACE_AMBIENT_COLOUR")
        vp:setAutoConstantInt("surf_emissive","SURFACE_EMISSIVE_COLOUR")
        vp:setAutoConstantInt("light_ambient","AMBIENT_LIGHT_COLOUR")

        local fp = prog(fname,lang,"FRAGMENT")
        fp.target = "ps_3_0"
        fp.sourceFile = "system/uber.shader"
        local tunit = 0
        if settings.shadow and settings.shadowReceive then
                for i=1,settings.shadowPCSSSplits do
                        defines = defines .. ";SEM_TEX_SHADOW"..i.."=SEM_TEX_"..tunit
                        tunit = tunit+1
                end
        end
        if stipple then
                defines = defines .. ";SEM_TEX_STIPPLE=SEM_TEX_"..tunit
                tunit = tunit+1
        end
        if diffuse_map then
                defines = defines .. ";SEM_TEX_DIFFUSE=SEM_TEX_"..tunit
                tunit = tunit+1
        end
        if normal_map then
                defines = defines .. ";SEM_TEX_NORMAL=SEM_TEX_"..tunit
                tunit = tunit+1
        end
        if spec_map then
                defines = defines .. ";SEM_TEX_SPECULAR=SEM_TEX_"..tunit
                tunit = tunit+1
        end
        fp.defines = defines..";FRAGMENT_PROGRAM=1"
        fp.entryPoint = "main"
        fp:reload()
        if not d3d9 then
                tunit = 0
                if settings.shadow and settings.shadowReceive then
                        for i=1,settings.shadowPCSSSplits do
                                fp:setConstantInt("shadow_map"..i, tunit);
                                tunit = tunit+1
                        end
                end
                if stipple then
                        fp:setConstantInt("stipple_map", tunit);
                        tunit = tunit+1
                end
                if diffuse_map then
                        fp:setConstantInt("diffuse_map", tunit);
                        tunit = tunit+1
                end
                if normal_map then
                        fp:setConstantInt("normal_map", tunit);
                        tunit = tunit+1
                end
                if spec_map then
                        fp:setConstantInt("spec_map", tunit);
                        tunit = tunit+1
                end
        end
        if settings.fog then
        fp:setAutoConstantInt("the_fog_params", "FOG_PARAMS")
        fp:setAutoConstantInt("the_fog_colour", "FOG_COLOUR")
        end
        fp:setAutoConstantInt("visibility", "CUSTOM", 0)
        fp:setConstantFloat("alpha_rej", 0)
        fp:setAutoConstantInt("light_diffuse", "DERIVED_LIGHT_DIFFUSE_COLOUR", 0)
        fp:setAutoConstantInt("light_specular", "DERIVED_LIGHT_SPECULAR_COLOUR", 0)
        fp:setAutoConstantInt("surf_diffuse", "SURFACE_DIFFUSE_COLOUR")
        fp:setAutoConstantInt("surf_specular", "SURFACE_SPECULAR_COLOUR")
        fp:setAutoConstantInt("surf_shininess", "SURFACE_SHININESS")


        return vp, fp
end

local function material_ff_internal (name, original, shadow_mat, tab)

        local mat = get_material(name)
        if mat==nil then
                mat = Material(name)
        else
                mat:removeAllTechniques()
                mat:createTechnique()
                mat:createPass(0)
        end

        mat:setAmbientVertex(0,0,true)

        if tab.backfaces then
                mat:setCull(0,0,false)
                mat:setManualCull(0,0,false)
        end     


        local tex_index = 0

        -- diffuse map
        if tab.diffuseMap then
                mat:createTextureUnitState(0,0)
                mat:setTextureName(0,0,tex_index, process_tex_name(original,tab.diffuseMap))
                set_texture_filtering(mat, tex_index)
                tex_index = tex_index + 1
        end

        -- diffuse colour
        local r,g,b = get_colour(tab.diffuseColour,1,1,1)
        mat:setDiffuse(0,0, r,g,b,1)

        r,g,b = get_colour(tab.ambientColour,r,g,b)
        mat:setAmbient(0,0, r,g,b)

        -- specular colour (default to white if specular map present)
        if tab.specularMap then
                r,g,b = get_colour(tab.specularColour,1,1,1)
        else
                r,g,b = get_colour(tab.specularColour,0,0,0)
        end
        mat:setSpecular(0,0, r,g,b,1,tab.specularity or 10)

        -- alpha stuff
        if tab.alpha then
                mat:setSceneBlending(0,0,"SRC_ALPHA","ONE_MINUS_SRC_ALPHA")
        end
        if tab.alpha then
                mat:setDepthWriteEnabled(0,0,false)
        end
        if tab.zsort then
                mat:setTransparentSortingForced(0,0,true)
        end

        return mat
end

local function material_internal (name, original, shadow_mat, fade, tab)

        local shf, shv = shader_names(tab.diffuseMap, tab.normalMap, tab.specularMap, tab.stipple~=false)

        local mat = get_material(name)
        if mat==nil then
                mat = Material(name)
        else
                mat:removeAllTechniques()
                mat:createTechnique()
                mat:createPass(0)
        end

        mat:setAmbientVertex(0,0,true)
        --mat:setDiffuseVertex(0,0,true)
        --mat:setSpecularVertex(0,0,true)
        --mat:setEmissiveVertex(0,0,true)


        mat:setShadowCasterMaterial(0,shadow_mat)

        mat:setFragmentProgram(0,0,shf)
        mat:setVertexProgram(0,0,shv)

        mat:setFragmentConstantFloat(0,0,"alpha_rej", tab.alphaReject or 63/255)

        if tab.backfaces then
                mat:setCull(0,0,false)
                mat:setManualCull(0,0,false)
        end     


        local tex_index = 0

        -- shadow maps
        if settings.shadow and settings.shadowReceive then
                for i=1,settings.shadowPCSSSplits do
                        mat:createTextureUnitState(0,0)
                        mat:setIsShadow(0,0,tex_index, true)
                        mat:setTextureFiltering(0,0,tex_index,"POINT","POINT","NONE")
                        tex_index = tex_index + 1
                end
        end

        -- stipple texture
        if tab.stipple ~= false then
                mat:createTextureUnitState(0,0)
                mat:setTextureName(0,0,tex_index, "system/stipple.bmp")
                mat:setTextureFiltering(0,0,tex_index,"POINT","POINT","NONE")
                tex_index = tex_index + 1
        end

        -- diffuse map
        if tab.diffuseMap then
                mat:createTextureUnitState(0,0)
                mat:setTextureName(0,0,tex_index, process_tex_name(original,tab.diffuseMap))
                set_texture_filtering(mat, tex_index)
                tex_index = tex_index + 1
        end

        -- normal map
        if tab.normalMap then
                mat:createTextureUnitState(0,0)
                mat:setTextureName(0,0,tex_index, process_tex_name(original,tab.normalMap))
                set_texture_filtering(mat, tex_index)
                tex_index = tex_index + 1
        end

        -- spec map
        if tab.specularMap then
                mat:createTextureUnitState(0,0)
                mat:setTextureName(0,0,tex_index, process_tex_name(original,tab.specularMap))
                set_texture_filtering(mat, tex_index)
                tex_index = tex_index + 1
        end

        -- diffuse colour
        local r,g,b = get_colour(tab.diffuseColour,1,1,1)
        mat:setDiffuse(0,0, r,g,b,1)

        r,g,b = get_colour(tab.ambientColour,r,g,b)
        mat:setAmbient(0,0, r,g,b)

        -- specular colour (default to white if specular map present)
        if tab.specularMap then
                r,g,b = get_colour(tab.specularColour,1,1,1)
        else
                r,g,b = get_colour(tab.specularColour,0,0,0)
        end
        mat:setSpecular(0,0, r,g,b,1,tab.specularity or 10)

        -- alpha stuff
        if tab.alpha or fade then
                mat:setSceneBlending(0,0,"SRC_ALPHA","ONE_MINUS_SRC_ALPHA")
        end
        if tab.alpha then
                mat:setDepthWriteEnabled(0,0,false)
        end
        if tab.zsort or fade then
                mat:setTransparentSortingForced(0,0,true)
        end

        mat.transparencyCastsShadows = tab.hasCaster or tab.stipple == false

        return mat
end

-- curry to avoid parentheses
local function material3(name,tab)
        tab.hasCaster = tab.alpha and tab.diffuseMap and settings.shadowUseDiffuse
        if settings.diffuseMaps == false then tab.diffuseMap = nil end
        if settings.normalMaps == false then tab.normalMap = nil end
        if settings.specularMaps == false then tab.specularMap = nil end
        local caster = tab.hasCaster and material_caster(name.."!", name, tab) or default_caster
        if settings.fixedFunction then
                material_ff_internal(name, name, caster, tab)
        else
                if tab.stipple == false then
                        material_internal(name.."'", name, caster, true, tab)
                end
                material_internal(name, name, caster, false, tab)
        end
end
local the_name = nil
local function material2(tab)
        local name = the_name
        tab = tab or {}
        --tab.stipple = false
        register_material(name,tab)
        material3(name,tab)
end
function material(name) the_name = name ; return material2 end


-- make sure global variable exists, to avoid error later one
default_caster = default_caster or nil


function do_reset_shaders ()
        local caster = make_program_caster_cg
        local receiver = make_program_cg

        caster(false)
        caster(true)

        -- always rebuild default caster
        material_caster("default_caster", "default_caster", {})

        receiver(true,true,false,false)
        receiver(true,true,false,true)
        receiver(true,true,true,false)
        receiver(true,true,true,true)
        receiver(true,false,true,false)
        receiver(true,false,true,true)
        receiver(true,false,false,false)
        receiver(true,false,false,true)
        receiver(false,true,false,false)
        receiver(false,true,false,true)
        receiver(false,true,true,false)
        receiver(false,true,true,true)
        receiver(false,false,false,false)
        receiver(false,false,false,true)
        receiver(false,false,true,false)
        receiver(false,false,true,true)

        -- no need to update pointer every time though
        default_caster = default_caster or get_material("default_caster")

end

function do_reset_materials()
        --regenerate_materials()
        material "BaseWhite" { }
        reprocess_all_registered_materials(material3)
end



-- vim: ts=8:sw=8:et
