-- (c) David Cunningham 2009, Licensed under the MIT license: http://www.opensource.org/licenses/mit-license.php

windowed_width = windowed_width or 800
windowed_height = windowed_height or 600

local function maybe_commit (self)
    if not self.c.autoUpdate then return end

    local reset_shadowmaps = false
    local reset_pcss = false
    local reset_shadows = false
    local reset_debug_shadow = false
    local reset_shaders = false
    local reset_materials = false
    local reset_fullscreen = false

    for k,v in pairs(self.p) do
        if self.c[k] ~= v then
            self.c[k] = v
    
            if k == "shadow" then
                reset_shaders = true
                reset_materials = true
            elseif k == "shadowRes" then
                reset_shadowmaps = true
                reset_shaders = true
            elseif k == "shadowBackwards" then
                sm.shadowCasterRenderBackFaces = v
            elseif k == "shadowCast" then
                reset_shadows = true
            elseif k == "shadowReceive" then
                reset_shaders = true
                reset_materials = true
            elseif k == "shadowUseDiffuse" then
                reset_materials = true
            elseif k == "shadowPCFBilinear" then
                reset_shaders = true
            elseif k == "shadowPCFTaps" then
                reset_shaders = true
            elseif k == "shadowPCFSpread" then
                reset_shaders = true
            elseif k == "shadowPCFEdgeDither" then
                reset_shaders = true
            elseif k == "shadowPCSSPadding" then
                reset_shaders = true
                reset_pcss = true
            elseif k == "shadowPCSSSplits" then
                reset_pcss = true
                reset_shaders = true
                reset_materials = true
                reset_shadowmaps = true
            elseif k == "shadowPCSSStart" then
                reset_pcss = true
            elseif k == "shadowPCSSEnd0" then
                reset_shaders = true
                reset_pcss = true
            elseif k == "shadowPCSSEnd1" then
                reset_shaders = true
                reset_pcss = true
            elseif k == "shadowPCSSEnd2" then
                reset_shaders = true
                reset_pcss = true
            elseif k == "shadowPCSSEnd3" then
                reset_shaders = true
                reset_pcss = true
            elseif k == "shadowPCSSAdj0" then
                reset_pcss = true
            elseif k == "shadowPCSSAdj1" then
                reset_pcss = true
            elseif k == "shadowPCSSAdj2" then
                reset_pcss = true
            elseif k == "shadowPCSSAdj3" then
                reset_pcss = true
            elseif k == "shadowPCSSSpreadFactor0" then
                reset_shaders = true
            elseif k == "shadowPCSSSpreadFactor1" then
                reset_shaders = true
            elseif k == "shadowPCSSSpreadFactor2" then
                reset_shaders = true
            elseif k == "shadowPCSSSpreadFactor3" then
                reset_shaders = true
            elseif k == "shadowFadeStart" then
                reset_shaders = true

            elseif k == "FOV" then
                gfx.cam.FOV = v
            elseif k == "res" then
                reset_fullscreen = true
            elseif k == "fullscreen" then
                reset_fullscreen = true
            elseif k == "farClip" then
                gfx.cam.farClipDistance = v
            elseif k == "visibility" then
                streamer.visibility = v
            elseif k == "noVertexProcessing" then
                reset_shaders = true
            elseif k == "noFragmentProcessing" then
                reset_shaders = true
            elseif k == "graphicsRAM" then
                set_texture_budget(v)
                set_mesh_budget(0)
            elseif k == "lockMemory" then
                if v then mlockall() else munlockall() end

            elseif k == "polygonMode" then
                gfx.cam.polygonMode = self.p.polygonMode
            elseif k == "normalMaps" then
                reset_materials = true
            elseif k == "diffuseMaps" then
                reset_materials = true
            elseif k == "specularMaps" then
                reset_materials = true
            elseif k == "vertexColours" then
                reset_shaders = true
            elseif k == "anisotropy" then
                reset_materials = true
            elseif k == "filtering" then
                reset_materials = true
            elseif k == "fog" then
                reset_shaders = true
            elseif k == "boundingBoxes" then
                sm.showBoundingBoxes = v
            elseif k == "fixedFunction" then
                reset_materials = true
            elseif k == "debugShadow" then
                reset_debug_shadow = true
            elseif k == "debugShadowSize" then
                reset_debug_shadow = true
            else
                error("Unexpected: "..k)
            end
        end
    end

    if reset_shadowmaps then
        -- may have to turn shadows off, too
        sm:setShadowTextureCountPerLightType("DIRECTIONAL",settings.shadowPCSSSplits)
        sm:setShadowTextureSettings(settings.shadowRes, settings.shadowPCSSSplits, "PF_FLOAT32_R");
    end

    if reset_pcss then
        local tab = {
                splits = settings.shadowPCSSSplits,
                padding = settings.shadowPCSSPadding,
                boundary0 = settings.shadowPCSSStart
        }
        for i=0,tab.splits-1 do
            tab["boundary"..(i+1)] = settings["shadowPCSSEnd"..i]
        end
        for i=0,tab.splits-1 do
            tab["optimalAdjustFactor"..i] = settings["shadowPCSSAdj"..i]
        end
        sm:setShadowCameraSetup("PSSM",tab)
    end

    if reset_debug_shadow then
        if settings.debugShadow then
            gfx.debugTexturePane.material:setTextureName(0,0,0,sm:getShadowTexture(settings.debugShadow).name)
            gfx.debugTexturePane.visible = true
        else
            gfx.debugTexturePane.visible = false
        end
        local sz = settings.debugShadowSize
        gfx.debugTexturePane.resize = Hud.LEFT(-1,sz,sz)
    end

    if reset_shadows then
            sm.shadowTechnique = (settings.shadow and settings.shadowCast) and "TEXTURE_MODULATIVE_INTEGRATED" or "NONE"
    end

    if reset_shaders then
        do_reset_shaders()
    end

    if reset_materials then
        do_reset_materials()
    end

    if reset_fullscreen then
        if settings.fullscreen then
            if get_main_win().isFullScreen then
                if get_main_win().width == settings.res[1] and get_main_win().height == settings.res[2] then
                    -- nothing to do
                else
                    -- change res, keep fullscreen
                    get_main_win():setFullScreen(true, settings.res[1], settings.res[2])
                end
            else
                -- make fullscreen
                windowed_width = get_main_win().width
                windowed_height = get_main_win().height
                get_main_win():setFullScreen(true, settings.res[1], settings.res[2])
            end
        else
            if get_main_win().isFullScreen then
                -- make windowed
                get_main_win():setFullScreen(false, windowed_width, windowed_height)
            else
                -- changing res does you no good if you're windowed
            end
        end
    end
    
end

local function change (self, k, v)
    if k == "autoUpdate" then
        ensure_one_of(v,{false,true})
        self.c[k] = v
    elseif k == "shadow" then
        ensure_one_of(v,{false,true})
        self.p[k] = v
    elseif k == "shadowBackwards" then
        ensure_one_of(v,{false,true})
        self.p[k] = v
    elseif k == "shadowRes" then
        ensure_one_of(v,{128,256,512,1024,2048,4096})
        self.p[k] = v
    elseif k == "shadowCast" then
        ensure_one_of(v,{false,true})
        self.p[k] = v
    elseif k == "shadowReceive" then
        ensure_one_of(v,{false,true})
        self.p[k] = v
    elseif k == "shadowUseDiffuse" then
        ensure_one_of(v,{false,true})
        self.p[k] = v
    elseif k == "shadowPCFBilinear" then
        ensure_one_of(v,{false,true})
        self.p[k] = v
    elseif k == "shadowPCFTaps" then
        ensure_int_range(v,1,6)
        self.p[k] = v
    elseif k == "shadowPCFSpread" then
        ensure_range(v,0,40)
        self.p[k] = v
    elseif k == "shadowPCFEdgeDither" then
        ensure_one_of(v,{false,true})
        self.p[k] = v
    elseif k == "shadowPCSSPadding" then
        ensure_range(v,0,100)
        self.p[k] = v
    elseif k == "shadowPCSSSplits" then
        ensure_int_range(v,1,4)
        self.p[k] = v
    elseif k == "shadowPCSSStart" then
        ensure_range(v,0,10000)
        self.p[k] = v
    elseif k == "shadowPCSSEnd0" then
        ensure_range(v,0,10000)
        self.p[k] = v
    elseif k == "shadowPCSSEnd1" then
        ensure_range(v,0,10000)
        self.p[k] = v
    elseif k == "shadowPCSSEnd2" then
        ensure_range(v,0,10000)
        self.p[k] = v
    elseif k == "shadowPCSSEnd3" then
        ensure_range(v,0,10000)
        self.p[k] = v
    elseif k == "shadowPCSSAdj0" then
        ensure_range(v,0,10000)
        self.p[k] = v
    elseif k == "shadowPCSSAdj1" then
        ensure_range(v,0,10000)
        self.p[k] = v
    elseif k == "shadowPCSSAdj2" then
        ensure_range(v,0,10000)
        self.p[k] = v
    elseif k == "shadowPCSSAdj3" then
        ensure_range(v,0,10000)
        self.p[k] = v
    elseif k == "shadowPCSSSpreadFactor0" then
        ensure_range(v,0,20)
        self.p[k] = v
    elseif k == "shadowPCSSSpreadFactor1" then
        ensure_range(v,0,20)
        self.p[k] = v
    elseif k == "shadowPCSSSpreadFactor2" then
        ensure_range(v,0,20)
        self.p[k] = v
    elseif k == "shadowPCSSSpreadFactor3" then
        ensure_range(v,0,20)
        self.p[k] = v
    elseif k == "shadowFadeStart" then
        ensure_range(v,0,10000)
        self.p[k] = v

    elseif k == "FOV" then
        ensure_range(v,0,120)
        self.p[k] = v
    elseif k == "res" then
        ensure_array(v,2)
        ensure_int_range(v[1],1,4096)
        ensure_int_range(v[2],1,4096)
        self.p[k] = v
    elseif k == "fullscreen" then
        ensure_one_of(v,{false,true})
        self.p[k] = v
    elseif k == "farClip" then
        ensure_range(v,1,10000)
        self.p[k] = v
    elseif k == "visibility" then
        ensure_range(v,0,5)
        self.p[k] = v
    elseif k == "noVertexProcessing" then
        ensure_one_of(v,{false,true})
        self.p[k] = v
    elseif k == "noFragmentProcessing" then
        ensure_one_of(v,{false,true})
        self.p[k] = v
    elseif k == "graphicsRAM" then
        ensure_int_range(v,0,1024)
        self.p[k] = v
    elseif k == "lockMemory" then
        ensure_one_of(v,{false,true})
        self.p[k] = v

    elseif k == "polygonMode" then
        ensure_one_of(v,{"SOLID","WIREFRAME","POINTS"})
        self.p[k] = v
    elseif k == "normalMaps" then
        ensure_one_of(v,{false,true})
        self.p[k] = v
    elseif k == "diffuseMaps" then
        ensure_one_of(v,{false,true})
        self.p[k] = v
    elseif k == "specularMaps" then
        ensure_one_of(v,{false,true})
        self.p[k] = v
    elseif k == "vertexColours" then
        ensure_one_of(v,{false,true})
        self.p[k] = v
    elseif k == "anisotropy" then
        ensure_int_range(v,0,16)
        self.p[k] = v
    elseif k == "filtering" then
        ensure_one_of(v,{false,true})
        self.p[k] = v
    elseif k == "fog" then
        ensure_one_of(v,{false,true})
        self.p[k] = v
    elseif k == "boundingBoxes" then
        ensure_one_of(v,{false,true})
        self.p[k] = v
    elseif k == "fixedFunction" then
        ensure_one_of(v,{false,true})
        self.p[k] = v
    elseif k == "debugShadow" then
        ensure_one_of(v,{0,1,2,3,false})
        self.p[k] = v
    elseif k == "debugShadowSize" then
        ensure_int_range(v,1,1024)
        self.p[k] = v

    else
        error("Unrecognised setting: "..tostring(k))
    end

    maybe_commit(self)
end

if not settings then
    settings = {
        -- current values
        c = {
            autoUpdate = false;
        },

        -- proposed values
        p = {
            shadow = true;
            shadowBackwards = true;
            shadowRes = 1024;
            shadowCast = true;
            shadowReceive = true;
            shadowUseDiffuse = true;
            shadowPCFBilinear = true;
            shadowPCFTaps = 2;
            shadowPCFSpread = 4;
            shadowPCFEdgeDither = true;
            shadowPCSSPadding = 0.6;
            shadowPCSSSplits = 3;
            shadowPCSSStart = 0.2;
            shadowPCSSEnd0 = 20;
            shadowPCSSEnd1 = 50;
            shadowPCSSEnd2 = 200;
            shadowPCSSEnd3 = 0;
            shadowPCSSAdj0 = 3;
            shadowPCSSAdj1 = 1;
            shadowPCSSAdj2 = 1;
            shadowPCSSAdj3 = 1;
            shadowPCSSSpreadFactor0 = 1;
            shadowPCSSSpreadFactor1 = 1;
            shadowPCSSSpreadFactor2 = 0.28;
            shadowPCSSSpreadFactor3 = 0.28;
            shadowFadeStart = 150;
            FOV = 55;
            fullscreen = get_main_win().isFullScreen;
            res = get_main_win().isFullScreen and {get_main_win().width, get_main_win().height} or {800,600};
            farClip = 1000;
            visibility = 1;
            noVertexProcessing = false;
            noFragmentProcessing = false;
            graphicsRAM = 128;
            lockMemory = true;
            normalMaps = true;
            diffuseMaps = true;
            specularMaps = true;
            vertexColours = true;
            anisotropy = 4;
            filtering = true;
            fog = true;
            boundingBoxes = false;
            fixedFunction = false;
            debugShadow = false;
            debugShadowSize = 128;
        }
    }
    for k,v in pairs(settings.p) do
        if user_cfg and user_cfg[k] ~= nil then
            settings.p[k] = user_cfg[k]
        end
    end
end

setmetatable(settings, {
    __index = function (self, k)
        local v = self.c[k] ;
        if v == nil then
            error("No such setting: \""..k.."\"",2)
        end
        return v
    end,
    __newindex = function (self, k, v) change(self,k,v) end
})

settings.autoUpdate = true
