// (c) David Cunningham 2009, Licensed under the MIT license: http://www.opensource.org/licenses/mit-license.php

#ifdef GRIT_HLSL
#define UNIFORM
#define SEP ,
#define LSEP
#define SEM_TEX_0 : TEXUNIT0
#define SEM_TEX_1 : TEXUNIT1
#define SEM_TEX_2 : TEXUNIT2
#define SEM_TEX_3 : TEXUNIT3
#define SEM_TEX_4 : TEXUNIT4
#define SEM_TEX_5 : TEXUNIT5
#define SEM_TEX_6 : TEXUNIT6
#define SEM_TEX_7 : TEXUNIT7
#define SEM_POSITION : POSITION
#define SEM_TANGENT : TANGENT0
#define SEM_NORMAL : NORMAL
#define SEM_COLOR : COLOR
#define SEM_COLOR0 : COLOR0
#define SEM_COLOR1 : COLOR1
#define SEM_TEXCOORD0 : TEXCOORD0
#define SEM_TEXCOORD1 : TEXCOORD1
#define SEM_TEXCOORD2 : TEXCOORD2
#define SEM_TEXCOORD3 : TEXCOORD3
#define SEM_TEXCOORD4 : TEXCOORD4
#define SEM_TEXCOORD5 : TEXCOORD5
#define SEM_TEXCOORD6 : TEXCOORD6
#define SEM_TEXCOORD7 : TEXCOORD7
#define SEM_TEXCOORD8 : TEXCOORD8
#define SEM_TEXCOORD9 : TEXCOORD9
#define SEM_TEXCOORD10 : TEXCOORD10
#define SEM_TEXCOORD11 : TEXCOORD11
#define PROGRAM_DEFINITION_BEGINS void main (
#define PROGRAM_DEFINITION_ENDS ) {
float4 mul4(float4x4 mat, float4 v) { return mul(mat,v); }
float3 mul3(float3x3 mat, float3 v) { return mul(mat,v); }
float saturate(float x) { return clamp(x,0.0,1.0); }
float3x3 rot_matrix(float3 X, float3 Y, float3 Z) { return transpose(float3x3(X,Y,Z)); }
#else
#version 130
#define float4 vec4
#define float3 vec3
#define float2 vec2
#define int4 ivec4
#define int3 ivec3
#define int2 ivec2
#define float3x3 mat3
#define float4x4 mat4
#define lerp mix
#define tex2D texture
#define UNIFORM uniform
#define SEM_TEX_0
#define SEM_TEX_1
#define SEM_TEX_2
#define SEM_TEX_3
#define SEM_TEX_4
#define SEM_TEX_5
#define SEM_TEX_6
#define SEM_TEX_7
#define SEM_POSITION
#define SEM_TANGENT
#define SEM_NORMAL
float4 mul4(float4x4 mat, float4 v) { return mat * v; }
float3 mul3(float3x3 mat, float3 v) { return mat * v; }
float saturate(float x) { return clamp(x,0.0,1.0); }
float3x3 rot_matrix(float3 X, float3 Y, float3 Z) { return float3x3(X,Y,Z); }
#define SEM_COLOR
#define SEM_COLOR0
#define SEM_COLOR1
#define SEM_TEXCOORD0
#define SEM_TEXCOORD1
#define SEM_TEXCOORD2
#define SEM_TEXCOORD3
#define SEM_TEXCOORD4
#define SEM_TEXCOORD5
#define SEM_TEXCOORD6
#define SEM_TEXCOORD7
#define SEM_TEXCOORD8
#define SEM_TEXCOORD9
#define SEM_TEXCOORD10
#define SEP ;
#define LSEP ;
#define PROGRAM_DEFINITION_BEGINS
#define PROGRAM_DEFINITION_ENDS void main(void) {
#endif

#define PI 3.141592653589

#define SHADOW_FACTOR 5000 // max distance from 0,0,0 to any point (in direction of sun)

#if NUM_SHADOW_MAPS >0

//#define SHADOW_BIAS1 SPREAD1/(-2.11+32.19*max(0.1,flat_light_exposure))
//#define SHADOW_BIAS2 SPREAD2/(-.537+10.355*max(0.1,flat_light_exposure))
//#define SHADOW_BIAS3 SPREAD3/(-0.187+3.533*max(flat_light_exposure,0.1))
#define SHADOW_BIAS1  -0.005//-SPREAD1*0.01
#define SHADOW_BIAS2  -0.005//-SPREAD2*0.01
#define SHADOW_BIAS3  -0.005//-SPREAD3*0.01

float tap_dist (sampler2D tex, float2 suv, float our_dist)
{
        return our_dist - tex2D(tex, suv).r;
}

float tap_test (sampler2D tex, float2 suv, float our_dist)
{
        return tap_dist(tex,suv,our_dist)>0 ? 1.0 : 0.0;
}

float bilinear_tap_test (sampler2D tex, int sz, float2 suv, float d)
{
        #if USE_BILINEAR_PCF_TAP == 0
        return tap_test(tex,suv,d);
        #else
        suv *= sz;
        float2 wh = floor(suv); // whole component


        float2 fr = suv - wh; // fractional component
        return lerp(lerp(tap_test(tex,(wh+float2(0,0))/sz,d), tap_test(tex,(wh+float2(0,1))/sz,d), fr.y),
                    lerp(tap_test(tex,(wh+float2(1,0))/sz,d), tap_test(tex,(wh+float2(1,1))/sz,d), fr.y),
                    fr.x);
/*
        return max(max(tap_test(tex,(wh+float2(0,0))/sz,d), tap_test(tex,(wh+float2(0,1))/sz,d)),
                    max(tap_test(tex,(wh+float2(1,0))/sz,d), tap_test(tex,(wh+float2(1,1))/sz,d))
                    );
*/
        #endif
}

float test_shadow (float3 vertex_ls, sampler2D tex, float our_dist,
                   float spread, float bias, float2 screen_pos)
{
        vertex_ls.xy /= vertex_ls.z;
        if (vertex_ls.x < 0) return 0;
        if (vertex_ls.y < 0) return 0;
        if (vertex_ls.x > 1) return 0;
        if (vertex_ls.y > 1) return 0;

        our_dist -= bias;
        our_dist /= SHADOW_FACTOR;


        // range (0,1) on each axis, equal border around all edges
        float2 offset = (float2(int(screen_pos.x)%FAKE_PCF_SIDE+0.5,
                                int(screen_pos.y)%FAKE_PCF_SIDE+0.5)) / FAKE_PCF_SIDE;

        if (offset.x > 0.5) offset.y = 1 - offset.y;

        float total = 0;
        for (int y=0 ; y<PCF_SIDE ; ++y) {
                for (int x=0 ; x<PCF_SIDE ; ++x) {
                        float2 p = (float2(x,y) + offset)/PCF_SIDE - float2(0.5,0.5);
                        p *= spread / 1024.0;
                        float2 suv = vertex_ls.xy + p;
                        total += bilinear_tap_test(tex,1024,suv,our_dist);
                }
        }

        return total / PCF_SIDE / PCF_SIDE;
}

#endif // NUM_SHADOW_MAPS>0

#if USE_STIPPLE_TEXTURE==1
bool stipple (int x, int y, float v, sampler2D tex)
{
    v *= 16.0; // 16 possibilities
    float2 uv = float2(x,y);
    // uv points to top left square now
    uv.x += 8.0 * (int(v)%4);
    uv.y += 8.0 * int(v/4);
    return tex2D(tex,uv/32.0).r < 0.5;
}
#endif

#ifdef VERTEX_PROGRAM
UNIFORM float4x4 world;
UNIFORM float4x4 view_proj;
#if NUM_SHADOW_MAPS > 0
UNIFORM float4x4 shadow_world_view_proj1;
#endif
#if NUM_SHADOW_MAPS > 1
UNIFORM float4x4 shadow_world_view_proj2;
#endif
#if NUM_SHADOW_MAPS > 2
UNIFORM float4x4 shadow_world_view_proj3;
#endif
#if NUM_SHADOW_MAPS > 3
UNIFORM float4x4 shadow_world_view_proj4;
#endif
UNIFORM float3 camera_pos;
UNIFORM float4 light_pos_os;
UNIFORM float4 light_pos_ws;

UNIFORM float3 surf_ambient;
UNIFORM float3 surf_emissive;
UNIFORM float3 light_ambient;

PROGRAM_DEFINITION_BEGINS
        in float3 vertex SEM_POSITION SEP
        in float3 normal SEM_NORMAL SEP
        #if USE_NORMAL_MAP==1
        in float3 tangent SEM_TANGENT SEP
        #endif
        #if USE_DIFFUSE_MAP==1 || USE_NORMAL_MAP==1 || USE_SPECULAR_MAP==1
        in float2 uv0 SEM_TEXCOORD0 SEP
        #endif
        #if USE_VERTEX_COLOURS==1
        in float4 colour SEM_COLOR SEP
        #if USE_NIGHT_COLOURS==1
        in float4 uv1 SEM_TEXCOORD1 SEP
        #endif
        #endif
        #if NO_VERTEX_PROCESSING==0
        out float4 out00 SEM_TEXCOORD0 SEP
        out float4 out01 SEM_TEXCOORD1 SEP
        out float4 out02 SEM_TEXCOORD2 SEP
        out float4 out03 SEM_TEXCOORD3 SEP
        out float4 out04 SEM_TEXCOORD4 SEP
        out float4 out05 SEM_TEXCOORD5 SEP
        out float4 out06 SEM_TEXCOORD6 SEP
        out float4 out07 SEM_TEXCOORD7 SEP
        #endif
        //out float4 out08 SEM_TEXCOORD8 SEP
        //out float4 out09 SEM_TEXCOORD9 SEP
        //out float4 out10 SEM_TEXCOORD10 SEP
        #ifdef GRIT_HLSL
        out float4 gl_Position : POSITION SEP
        #endif
        out float4 ambient SEM_COLOR0 LSEP
PROGRAM_DEFINITION_ENDS
        #if NO_VERTEX_PROCESSING==1
        gl_Position = float4(0,0,-10,1);
        ambient = float4(1,1,1,1);
        return;
        #else

        #if USE_VERTEX_COLOURS==1
        ambient.rgb = surf_ambient * light_ambient * colour.rgb + surf_emissive;
        #else
        ambient.rgb = surf_ambient * light_ambient + surf_emissive;
        #endif
        ambient.a = 1;

        // do ambient lighting

        // lighting vectors for fragment program
        float3 vertex_to_light_ = light_pos_os.xyz;  // assume directional light
        float3 vertex_to_camera_ = normalize(camera_pos - vertex);
        float3 normal_ = normalize(normal);
        float3 tangent_ = float3(0,1,0);
        float3 binormal_ = float3(1,0,0);
        #if USE_NORMAL_MAP==1
        tangent_ = tangent;
        binormal_ = cross(normal,tangent);
        #endif

        float2 uv0_ = float2(0,0);
        #if USE_DIFFUSE_MAP==1 || USE_NORMAL_MAP==1 || USE_SPECULAR_MAP==1
        uv0_ = uv0;
        #endif

        // vertex position in light space for shadow mapping
        #if NUM_SHADOW_MAPS > 0
        float4 vertex_ls1 = mul4(shadow_world_view_proj1, float4(vertex,1));
        float3 vertex_ls1_ = vertex_ls1.xyw;
        #endif
        #if NUM_SHADOW_MAPS > 1
        float4 vertex_ls2 = mul4(shadow_world_view_proj2, float4(vertex,1));
        float3 vertex_ls2_ = vertex_ls2.xyw;
        #endif
        #if NUM_SHADOW_MAPS > 2
        float4 vertex_ls3 = mul4(shadow_world_view_proj3, float4(vertex,1));
        float3 vertex_ls3_ = vertex_ls3.xyw;
        #endif
        #if NUM_SHADOW_MAPS > 3
        float4 vertex_ls4 = mul4(shadow_world_view_proj4, float4(vertex,1));
        float3 vertex_ls4_ = vertex_ls4.xyw;
        #endif

        float3 vertex_ws = mul4(world, float4(vertex,1)).xyz;
        gl_Position = mul4(view_proj, float4(vertex_ws,1));
        float light_dist_ = -dot(vertex_ws,light_pos_ws.xyz);
        float cam_dist_ = gl_Position.z;

        out00.xyz = vertex_to_light_;
        out00.w = light_dist_;

        out01.xyz = vertex_to_camera_;
        out01.w = cam_dist_;

        out02.xyz = normal_;
        out02.w = uv0_.x;

        out03.xyz = tangent_;
        out03.w = uv0_.y;

        out04.xyz = binormal_;
        //out04.w = 0;

        #if NUM_SHADOW_MAPS > 0
        out05.xyz = vertex_ls1_;
        #endif
        #if NUM_SHADOW_MAPS > 1
        out06.xyz = vertex_ls2_;
        #endif
        #if NUM_SHADOW_MAPS > 2
        out07.xyz = vertex_ls3_;
        #endif
        #if NUM_SHADOW_MAPS > 3
        out05.w = vertex_ls4_.x;
        out06.w = vertex_ls4_.y;
        out07.w = vertex_ls4_.z;
        #endif

        #endif // NO_VERTEX_PROCESSING
}
#endif // VERTEX_PROGRAM


#ifdef FRAGMENT_PROGRAM
#if NUM_SHADOW_MAPS>0
UNIFORM sampler2D shadow_map1 SEM_TEX_SHADOW1;
#endif
#if NUM_SHADOW_MAPS>1
UNIFORM sampler2D shadow_map2 SEM_TEX_SHADOW2;
#endif
#if NUM_SHADOW_MAPS>2
UNIFORM sampler2D shadow_map3 SEM_TEX_SHADOW3;
#endif
#if NUM_SHADOW_MAPS>3
UNIFORM sampler2D shadow_map4 SEM_TEX_SHADOW4;
#endif
#if USE_STIPPLE_TEXTURE==1
UNIFORM sampler2D stipple_map SEM_TEX_STIPPLE;
#endif
#if USE_DIFFUSE_MAP==1
UNIFORM sampler2D diffuse_map SEM_TEX_DIFFUSE;
#endif
#if USE_NORMAL_MAP==1
UNIFORM sampler2D normal_map  SEM_TEX_NORMAL;
#endif
#if USE_SPECULAR_MAP==1
UNIFORM sampler2D spec_map    SEM_TEX_SPECULAR;
#endif
#if USE_FOG==1
UNIFORM float4 the_fog_params;
UNIFORM float3 the_fog_colour;
#endif
UNIFORM float4 visibility;
UNIFORM float alpha_rej;
UNIFORM float3 light_diffuse;
UNIFORM float3 light_specular;
UNIFORM float3 surf_diffuse;
UNIFORM float3 surf_specular;
UNIFORM float surf_shininess;

PROGRAM_DEFINITION_BEGINS
        in float4 ambient SEM_COLOR0 SEP
        in float4 out00 SEM_TEXCOORD0 SEP
        in float4 out01 SEM_TEXCOORD1 SEP
        in float4 out02 SEM_TEXCOORD2 SEP
        in float4 out03 SEM_TEXCOORD3 SEP
        in float4 out04 SEM_TEXCOORD4 SEP
        #if NUM_SHADOW_MAPS > 0
        in float4 out05 SEM_TEXCOORD5 SEP
        #endif
        #if NUM_SHADOW_MAPS > 1
        in float4 out06 SEM_TEXCOORD6 SEP
        #endif
        #if NUM_SHADOW_MAPS > 2
        in float4 out07 SEM_TEXCOORD7 SEP
        #endif
        //in float4 out08 SEM_TEXCOORD8 SEP
        //in float4 out09 SEM_TEXCOORD9 SEP
        //in float4 out10 SEM_TEXCOORD10 SEP
        #ifdef GRIT_HLSL
        in float2 gl_FragCoord : VPOS,
        #endif
        out float4 pixel SEM_COLOR LSEP
PROGRAM_DEFINITION_ENDS


        #if NO_FRAGMENT_PROCESSING == 1
        pixel = float4(1,1,0,1);
        return;
        #endif

        float3 vertex_to_light_ = out00.xyz;
        float light_dist_ = out00.w;

        float3 vertex_to_camera_ = out01.xyz;
        float cam_dist_ = out01.w;

        float3 normal_ = out02.xyz;
        float3 tangent_ = out03.xyz;
        float2 uv0_ = float2(out02.w, out03.w);

        float3 binormal_ = out04.xyz;

        #if NUM_SHADOW_MAPS > 0
        float3 vertex_ls1_ = out05.xyz;
        #endif
        #if NUM_SHADOW_MAPS > 1
        float3 vertex_ls2_ = out06.xyz;
        #endif
        #if NUM_SHADOW_MAPS > 2
        float3 vertex_ls3_ = out07.xyz;
        #endif
        #if NUM_SHADOW_MAPS > 3
        float3 vertex_ls4_ = float3(out05.x, out05.y, out05.z);
        #endif

        float2 screen_pos = gl_FragCoord.xy;

        // start with no lighting
        float3 dyn_light = float3(0,0,0);
        float3 spec_light = float3(0,0,0);

        float3 normal = normal_;
        float flat_light_exposure = max(0,dot(vertex_to_light_,normal));
        #if USE_NORMAL_MAP==1
        float3 nm_tap = float3(-2,2,2)*(tex2D(normal_map, uv0_).xyz - .5);
        normal = mul3(rot_matrix(tangent_, binormal_, normal_), nm_tap);
        #endif

        // compute diffuse lighting
        float light_exposure = max(0,dot(vertex_to_light_,normal));
        dyn_light += light_exposure * light_diffuse * surf_diffuse;

        // compute specular lighting
        float3 r = reflect(-vertex_to_light_,normal);
        spec_light = pow(max(0.00001,dot(r,vertex_to_camera_)),surf_shininess)
                     * light_specular * surf_specular;
        #if USE_SPECULAR_MAP==1
        spec_light *= tex2D(spec_map, uv0_).xyz;
        #endif


        // compute shadow (reduces dynamic lighting if in shadow)
        float shadowyness = 0;
        #if NUM_SHADOW_MAPS > 0
        if (cam_dist_ < SHADOW_DIST1) {
                shadowyness = test_shadow(vertex_ls1_, shadow_map1, light_dist_, SPREAD1 * light_exposure, SHADOW_BIAS1, screen_pos);
        }
        #if NUM_SHADOW_MAPS > 1
        else if (cam_dist_ < SHADOW_DIST2) {
                shadowyness = test_shadow(vertex_ls2_, shadow_map2, light_dist_, SPREAD2 * light_exposure, SHADOW_BIAS2, screen_pos);
        }
        #endif
        #if NUM_SHADOW_MAPS > 2
        else if (cam_dist_ < SHADOW_DIST3) {
                shadowyness = test_shadow(vertex_ls3_, shadow_map3, light_dist_, SPREAD3 * light_exposure, SHADOW_BIAS3, screen_pos);
        }
        #endif
        #if NUM_SHADOW_MAPS > 3
        else if (cam_dist_ < SHADOW_DIST4) {
                shadowyness = test_shadow(vertex_ls4_, shadow_map4, light_dist_, SPREAD4 * light_exposure, SHADOW_BIAS4, screen_pos);
        }
        #endif

        float fade = min(1, (SHADOW_FADE_END-cam_dist_)/(SHADOW_FADE_END-SHADOW_FADE_START));
        shadowyness *= fade;
        #endif

        //shadowyness = (min(shadowyness+0.4,.9) - 0.4) * 2;
        //shadowyness = min(shadowyness+0.4,.9);

        // shadow affects dynamic and specular light
        dyn_light.rgb -= dyn_light.rgb * shadowyness;
        spec_light.rgb -= spec_light.rgb * shadowyness;


        // Ambient as built up using scene, material ambient, and vertex
        // colours.  Ambient is a way of cheating yourself more light to get
        // out of darkness, so if you could have just as well gotten that light
        // legitimately then it doesn't make a difference.
        pixel.rgb = max(ambient.rgb, dyn_light);
        pixel.a = ambient.a;

        #if USE_DIFFUSE_MAP==1
        float4 texel = tex2D(diffuse_map, uv0_);
        //pixel *= float4((normal_ + float3(1,1,1))/2,1);
        pixel *= texel;
        //pixel.rgb = float3(uv0_,0);
        #endif

        // Add on spec, this will saturate in some cases giving the 'glare' effect we want.
        pixel.rgb += spec_light;


        // cull pixels after lighting and texture, but before fading
        if (pixel.a <= alpha_rej) discard;

        // exp2 fog
        #if USE_FOG==1
        float clarity = the_fog_params.x * cam_dist_;
        clarity = clamp(exp(-clarity * clarity),0.0,1.0);
        pixel.rgb = lerp(the_fog_colour, pixel.rgb, clarity);
        #endif


        float v = visibility.x;
        #if USE_STIPPLE_TEXTURE==1
        if (v<1) {

                int x = (int(screen_pos.x) % 8);
                int y = (int(screen_pos.y) % 8);
                
                if (stipple(x,y,v,stipple_map)) discard;
        }
        #else
                pixel.a *= v;
        #endif
}
#endif // FRAGMENT_PROGRAM







#ifdef CASTER_VERTEX_PROGRAM
UNIFORM float4x4 world;
UNIFORM float4x4 view_proj;
UNIFORM float4 light_pos_ws;

PROGRAM_DEFINITION_BEGINS
        in float3 vertex SEM_POSITION SEP
        #if USE_DIFFUSE_MAP==1
        in float2 uv0 SEM_TEXCOORD0 SEP
        #endif
        out float4 out00 SEM_TEXCOORD0 SEP
        #ifdef GRIT_HLSL
        out float4 gl_Position : POSITION
        #endif
PROGRAM_DEFINITION_ENDS

        float2 uv0_ = float2(0,0);
        #if USE_DIFFUSE_MAP==1
        uv0_ = uv0;
        #endif

        float3 vertex_ws = mul4(world, float4(vertex,1)).xyz;
        gl_Position = mul4(view_proj, float4(vertex_ws,1));
        float light_dist_ = -dot(vertex_ws,light_pos_ws.xyz);

        out00.xy = uv0_;
        out00.z = light_dist_;
}
#endif // CASTER_VERTEX_PROGRAM

#ifdef CASTER_FRAGMENT_PROGRAM
#if USE_DIFFUSE_MAP==1
UNIFORM sampler2D diffuse_map SEM_TEX_DIFFUSE;
#endif
#if USE_STIPPLE_TEXTURE==1
UNIFORM sampler2D stipple_map SEM_TEX_STIPPLE;
UNIFORM float4 visibility;
#endif
UNIFORM float alpha_rej;

PROGRAM_DEFINITION_BEGINS
        in float4 out00 SEM_TEXCOORD0 SEP
        #ifdef GRIT_HLSL
        in float4 gl_FragCoord : VPOS,
        #endif
        out float4 pixel SEM_COLOR LSEP
PROGRAM_DEFINITION_ENDS

        float2 uv0_ = out00.xy;
        float light_dist_ = out00.z;
        float2 screen_pos = gl_FragCoord.xy;

        float alpha = 1;

        #if USE_DIFFUSE_MAP==1
        alpha *= tex2D(diffuse_map, uv0_).a;
        #endif

        if (alpha <= alpha_rej) discard;

        #if USE_STIPPLE_TEXTURE==1
        float v = visibility.x;

        if (v<1) {
                int x = int(screen_pos.x) % 8;
                int y = int(screen_pos.y) % 8;
                if (stipple(x,y,v,stipple_map)) discard;
        }
        #endif


        pixel.rgba = float4(light_dist_ / SHADOW_FACTOR,0,0,1);

}
#endif // CASTER_FRAGMENT_PROGRAM

// vim: ts=8:sw=8:et
