-- (c) David Cunningham 2009, Licensed under the MIT license: http://www.opensource.org/licenses/mit-license.php

print("Loading ui.lua")

BindTable = { }

function BindTable.new()
        local self = {
                binds = {},
                buttonsDown = {},
        }
        return make_instance(self,BindTable,function (k)
                if BindTable[k] then return false end
                return true, self.binds[k] or self.binds["+"..k]
        end)
end

function BindTable:bind(event, down, up, rep)
        self.binds["+"..event] = down
        self.binds["-"..event] = up
        if rep == true then
                rep = down
        end
        self.binds["="..event] = rep
end

function BindTable:notify(event)
        local func = self.binds[event]
        if func == nil then return true end
        xpcall(function()func(event)end, error_handler)
        return false
end

function BindTable:flush(key)
        if key == nil then
                for k,_ in pairs(self.buttonsDown) do
                        local event = "-"..k
                        self:notify(event)
                end
                self.buttonsDown = {}
        else
                if self.buttonsDown[key] then
                        local event = "-"..key
                        self:notify(event)
                        self.buttonsDown[key] = nil
                end
        end
end

function BindTable:process(event)
        local c = event:sub(1,1)
        if c == "+" then
                self.buttonsDown[event:sub(2)] = true
        elseif c == "-" then
                self.buttonsDown[event:sub(2)] = nil
        end
        return self:notify(event)
end


UI = { }

function UI.new()
        local self = {
                buttonsDown = {},
                pressCallbacks = CallbackReg.new(),
                binds = BindTable.new(),
                pointerGrabCallbacks = CallbackReg.new(),
                pointerCallbacks = CallbackReg.new(),
                priorGrabX =0, priorGrabY = 0,
                lastX = 0, lastY = 0,
                grabbed = false,
                shouldBeGrabbed = false,
        }

        -- event callbacks allow you to register a callback on a particular key or mouse event
        -- general callbacks give you everything (e.g. for text input)
        self.pressCallbacks:insert("UI.system", function (v)
                local c = v:sub(1,1)
                if c == "+" then
                        self.buttonsDown[v:sub(2)] = true
                elseif c == "-" then
                        self.buttonsDown[v:sub(2)] = nil
                end
        end)

        main.frameCallbacks:insert("UI.frameCallback",UI.frameCallback)

        return make_instance(self,UI)
end

function UI:destroy()
        main.frameCallbacks:removeByName("UI.frameCallback")
end


function UI:alt() return self.buttonsDown.Alt end
function UI:shift() return self.buttonsDown.Shift end
function UI:ctrl() return self.buttonsDown.Ctrl end

function UI:down(key) return self.buttonsDown[key] end
function UI:keys() return pairs(self.buttonsDown) end

function UI:bind(...) return self.binds:bind(...) end

function UI:flush()
        if player_ctrl ~= nil then
                player_ctrl:flush()
        end
        self.binds:flush()
end

function UI:mouse(key)
        key = key:sub(2)
        for _,key2 in ipairs{"left","right","middle","up","down"} do
                if key == key2 then return true end
                if key == key2 then return true end
        end
end


function UI:getGrab() return self.grabbed end
function UI:getShouldBeGrabbed() return self.shouldBeGrabbed end

function UI:grab()
        self.shouldBeGrabbed = true
        self:updateGrabbed()
end

function UI:ungrab()
        self.shouldBeGrabbed = false
        self:updateGrabbed()
end     

function UI:updateGrabbed()
        local should_be_grabbed = have_focus() and self.shouldBeGrabbed
        if should_be_grabbed == self.grabbed then return end
        if should_be_grabbed then
                self.priorGrabX = self.lastX
                self.priorGrabY = self.lastY 
                set_mouse_grab(true)
                set_mouse_hide(true)
        else
                set_mouse_grab(false)
                set_mouse_pos(self.priorGrabX,self.priorGrabY)
                set_mouse_hide(false)
        end
        self.grabbed = should_be_grabbed
end


function UI.frameCallback ()

        local presses = get_keyb_presses()
        local moved,buttons,x,y,rel_x,rel_y = get_mouse_events()

        for _,key in ipairs(presses) do
                if get_keyb_verbose() then echo("Lua key event: "..key) end
                if ui.pressCallbacks:execute(key) ~= false then
                        ui.binds:process(key)
                end
        end

        for _,button in ipairs(buttons) do
                if ui.pressCallbacks:execute(button) ~= false then
                        ui.binds:process(button)
                end
        end


        if moved then
                ui.lastX, ui.lastY = x, y
                ui.lastRelX, ui.lastRelY = rel_x, rel_y
                local callbacks = ui:getGrab() and ui.pointerGrabCallbacks or ui.pointerCallbacks
                callbacks:execute(rel_x, rel_y)
        end

end




if ui ~= nil then
        ui:destroy()
end
ui = UI.new()



-- vim: shiftwidth=8:tabstop=8:expandtab
